@@ -33,7 +33,8 @@ This example layer is blended using the "hard light" mode at 50% opacity:
 
     {"src": "hillshading", "mode": "hard light", "opacity": 0.5}
 
-Currently-supported blend modes include "screen", "multiply", and "hard light".
+Currently-supported blend modes include "screen", "multiply", "linear light",
+and "hard light".
 
 Layers can also be affected by adjustments. Adjustments are specified as an
 array of names and parameters. This example layer has been slightly darkened
@@ -42,7 +43,10 @@ to 50% gray while leaving black and white alone:
 
     {"src": "hillshading", "adjustments": [ ["curves", [0, 181, 255]] ]}
 
-Currently, only the "curves" adjustment is supported.
+Available adjustments:
+  "threshold" - apply_threshold_adjustment()
+  "curves" - apply_curves_adjustment()
+  "curves2" - apply_curves2_adjustment()
 
 Finally, the stacking feature allows layers to combined in more complex ways.
 This example stack combines a background color and foreground layer:
@@ -98,7 +102,7 @@ A complete example configuration might look like this:
         {
           "provider":
           {
-            "class": "TileStache.Goodies.Providers.Composite.Provider",
+            "class": "TileStache.Goodies.Providers.Composite:Provider",
             "kwargs":
             {
               "stack":
@@ -153,8 +157,8 @@ except ImportError:
 import TileStache
 
 try:
-    import sympy
     import numpy
+    import sympy
 except ImportError:
     # At least we can build the docs
     pass
@@ -212,11 +216,11 @@ class Provider:
         
     def renderTile(self, width, height, srs, coord):
     
-        image = Image.new('RGBA', (width, height), (0, 0, 0, 0))
+        rgba = [numpy.zeros((width, height), float) for chan in range(4)]
         
-        image = self.stack.render(self.layer.config, image, coord)
+        rgba = self.stack.render(self.layer.config, rgba, coord)
         
-        return image
+        return _rgba2img(rgba)
 
 class Composite(Provider):
     """ An old name for the Provider class, deprecated for the next version.
@@ -271,7 +275,7 @@ class Layer:
         self.blendmode = blendmode
         self.adjustments = adjustments
 
-    def render(self, config, input_img, coord):
+    def render(self, config, input_rgba, coord):
         """ Render this image layer.
 
             Given a configuration object, starting image, and coordinate,
@@ -279,7 +283,7 @@ class Layer:
         """
         has_layer, has_color, has_mask = False, False, False
         
-        output_rgba = _img2rgba(input_img)
+        output_rgba = [chan.copy() for chan in input_rgba]
     
         if self.layername:
             layer = config.layers[self.layername]
@@ -334,9 +338,7 @@ class Layer:
         else:
             raise KnownUnknown("You have to provide at least some combination of src, color and mask to Composite Layer")
 
-        output_img = _rgba2img(output_rgba)
-        
-        return output_img
+        return output_rgba
 
 class Stack:
     """ A stack of image layers.
@@ -351,28 +353,25 @@ class Stack:
         """
         self.layers = layers
 
-    def render(self, config, input_img, coord):
+    def render(self, config, input_rgba, coord):
         """ Render this image stack.
 
             Given a configuration object, starting image, and coordinate,
             return an output image with the results of all the layers in
             this stack pasted on in turn.
         """
-        stack_img = Image.new('RGBA', input_img.size, (0, 0, 0, 0))
+        stack_rgba = [numpy.zeros(chan.shape, chan.dtype) for chan in input_rgba]
         
         for layer in self.layers:
             try:
-                stack_img = layer.render(config, stack_img, coord)
+                stack_rgba = layer.render(config, stack_rgba, coord)
             except IOError:
                 # Be permissive of I/O errors getting sub-layers, for example if a
                 # proxy layer referenced here doesn't have an image for a zoom level.
                 # TODO: regret this later.
                 pass
 
-        output_img = input_img.copy()
-        output_img.paste(stack_img, (0, 0), stack_img)
-        
-        return output_img
+        return blend_images(input_rgba, stack_rgba[:3], stack_rgba[3], 1, None)
 
 def make_color(color):
     """ Convert colors expressed as HTML-style RGB(A) strings to tuples.
@@ -440,26 +439,76 @@ def apply_adjustments(rgba, adjustments):
     
         Working adjustments:
         
+          threshold:
+            Calls apply_threshold_adjustment()
+        
           curves:
             Calls apply_curves_adjustment()
+        
+          curves2:
+            Calls apply_curves2_adjustment()
     """
     if not adjustments:
         return rgba
 
-    for (name, args) in adjustments:
-        if name == 'curves':
+    for adjustment in adjustments:
+        name, args = adjustment[0], adjustment[1:]
+
+        if name == 'threshold':
+            rgba = apply_threshold_adjustment(rgba, *args)
+        
+        elif name == 'curves':
             rgba = apply_curves_adjustment(rgba, *args)
         
+        elif name == 'curves2':
+            rgba = apply_curves2_adjustment(rgba, *args)
+        
         else:
             raise KnownUnknown('Unrecognized composite adjustment: "%s" with args %s' % (name, repr(args)))
     
     return rgba
 
-def apply_curves_adjustment(rgba, black, grey, white):
-    """ *write me*
+def apply_threshold_adjustment(rgba, red_value, green_value=None, blue_value=None):
+    """
+    """
+    if green_value is None or blue_value is None:
+        # if there aren't three provided, use the one
+        green_value, blue_value = red_value, red_value
+
+    # channels
+    red, green, blue, alpha = rgba
+    
+    # knowns are given in 0-255 range, need to be converted to floats
+    red_value, green_value, blue_value = red_value / 255.0, green_value / 255.0, blue_value / 255.0
+    
+    red[red > red_value] = 1
+    red[red <= red_value] = 0
+    
+    green[green > green_value] = 1
+    green[green <= green_value] = 0
+    
+    blue[blue > blue_value] = 1
+    blue[blue <= blue_value] = 0
+    
+    return red, green, blue, alpha
+
+def apply_curves_adjustment(rgba, black_grey_white):
+    """ Adjustment inspired by Photoshop "Curves" feature.
+    
+        Arguments are three integers that are intended to be mapped to black,
+        grey, and white outputs. Curves2 offers more flexibility, see
+        apply_curves2_adjustment().
+        
+        Darken a light image by pushing light grey to 50% grey, 0xCC to 0x80:
+    
+          [
+            "curves",
+            [0, 204, 255]
+          ]
     """
     # channels
     red, green, blue, alpha = rgba
+    black, grey, white = black_grey_white
     
     # coefficients
     a, b, c = [sympy.Symbol(n) for n in 'abc']
@@ -478,17 +527,79 @@ def apply_curves_adjustment(rgba, black, grey, white):
     a, b, c = [float(co[n]) * numpy.ones(red.shape, numpy.float32) for n in (a, b, c)]
     
     # arithmetic
-    red   = a * red**2   + b * red   + c
-    green = a * green**2 + b * green + c
-    blue  = a * blue**2  + b * blue  + c
+    red   = numpy.clip(a * red**2   + b * red   + c, 0, 1)
+    green = numpy.clip(a * green**2 + b * green + c, 0, 1)
+    blue  = numpy.clip(a * blue**2  + b * blue  + c, 0, 1)
     
     return red, green, blue, alpha
 
+def apply_curves2_adjustment(rgba, map_red, map_green=None, map_blue=None):
+    """ Adjustment inspired by Photoshop "Curves" feature.
+    
+        Arguments are given in the form of three value mappings, typically
+        mapping black, grey and white input and output values. One argument
+        indicates an effect applicable to all channels, three arguments apply
+        effects to each channel separately.
+    
+        Simple monochrome inversion:
+    
+          [
+            "curves2",
+            [[0, 255], [128, 128], [255, 0]]
+          ]
+    
+        Darken a light image by pushing light grey down by 50%, 0x99 to 0x66:
+    
+          [
+            "curves2",
+            [[0, 255], [153, 102], [255, 0]]
+          ]
+    
+        Shaded hills, with Imhof-style purple-blue shadows and warm highlights: 
+        
+          [
+            "curves2",
+            [[0, 22], [128, 128], [255, 255]],
+            [[0, 29], [128, 128], [255, 255]],
+            [[0, 65], [128, 128], [255, 228]]
+          ]
+    """
+    if map_green is None or map_blue is None:
+        # if there aren't three provided, use the one
+        map_green, map_blue = map_red, map_red
+
+    # channels
+    red, green, blue, alpha = rgba
+    out = []
+    
+    for (chan, input) in ((red, map_red), (green, map_green), (blue, map_blue)):
+        # coefficients
+        a, b, c = [sympy.Symbol(n) for n in 'abc']
+        
+        # parameters given in 0-255 range, need to be converted to floats
+        (in_1, out_1), (in_2, out_2), (in_3, out_3) \
+            = [(in_ / 255.0, out_ / 255.0) for (in_, out_) in input]
+        
+        # quadratic function
+        eqs = [a * in_1**2 + b * in_1 + c - out_1,
+               a * in_2**2 + b * in_2 + c - out_2,
+               a * in_3**2 + b * in_3 + c - out_3]
+        
+        co = sympy.solve(eqs, a, b, c)
+        
+        # arrays for each coefficient
+        a, b, c = [float(co[n]) * numpy.ones(chan.shape, numpy.float32) for n in (a, b, c)]
+        
+        # arithmetic
+        out.append(numpy.clip(a * chan**2 + b * chan + c, 0, 1))
+    
+    return out + [alpha]
+
 def blend_images(bottom_rgba, top_rgb, mask_chan, opacity, blendmode):
     """ Blend images using a given mask, opacity, and blend mode.
     
         Working blend modes:
-        None for plain pass-through, "screen", "multiply", and "hard light".
+        None for plain pass-through, "screen", "multiply", "linear light", and "hard light".
     """
     if opacity == 0 or not mask_chan.any():
         # no-op for zero opacity or empty mask
@@ -504,6 +615,7 @@ def blend_images(bottom_rgba, top_rgb, mask_chan, opacity, blendmode):
     else:
         blend_functions = {'screen': blend_channels_screen,
                            'multiply': blend_channels_multiply,
+                           'linear light': blend_channels_linear_light,
                            'hard light': blend_channels_hard_light}
 
         if blendmode in blend_functions:
@@ -524,12 +636,25 @@ def blend_images(bottom_rgba, top_rgb, mask_chan, opacity, blendmode):
     if gr.any():
         # we have some shades of gray to take care of
         for c in (0, 1, 2):
-            output_rgba[c][gr] = (1 - mask_chan[gr]) * bottom_rgba[c][gr] \
-                                     + mask_chan[gr] * output_rgba[c][gr]
+            #
+            # Math borrowed from Wikipedia; C0 is the variable alpha_denom:
+            # http://en.wikipedia.org/wiki/Alpha_compositing#Analytical_derivation_of_the_over_operator
+            #
+            
+            alpha_denom = 1 - (1 - mask_chan) * (1 - bottom_rgba[3])
+            nz = alpha_denom > 0 # non-zero alpha denominator
+            
+            alpha_ratio = mask_chan[nz] / alpha_denom[nz]
+            
+            output_rgba[c][nz] = output_rgba[c][nz] * alpha_ratio \
+                               + bottom_rgba[c][nz] * (1 - alpha_ratio)
+            
+            # let the zeros perish
+            output_rgba[c][~nz] = 0
     
     # output mask is the screen of the existing and overlaid alphas
     blend_channels_screen(output_rgba[3], bottom_rgba[3], mask_chan)
-    
+
     return output_rgba
 
 def blend_channels_screen(output_chan, bottom_chan, top_chan):
@@ -546,6 +671,13 @@ def blend_channels_multiply(output_chan, bottom_chan, top_chan):
     """
     output_chan[:,:] = bottom_chan[:,:] * top_chan[:,:]
 
+def blend_channels_linear_light(output_chan, bottom_chan, top_chan):
+    """ Modify output channel in-place with the combination of bottom and top channels.
+    
+        Math from http://illusions.hu/effectwiki/doku.php?id=linear_light_blending
+    """
+    output_chan[:,:] = numpy.clip(bottom_chan[:,:] + 2 * top_chan[:,:] - 1, 0, 1)
+
 def blend_channels_hard_light(output_chan, bottom_chan, top_chan):
     """ Modify output channel in-place with the combination of bottom and top channels.
     
@@ -614,6 +746,15 @@ if __name__ == '__main__':
     import TileStache.Config
     import ModestMaps.Core
     
+    class SizelessImage:
+        """ Wrap an image without wrapping the size() method, for Layer.render().
+        """
+        def __init__(self, img):
+            self.img = img
+        
+        def save(self, out, format):
+            self.img.save(out, format)
+    
     class TinyBitmap:
         """ A minimal provider that only returns 3x3 bitmaps from strings.
         """
@@ -621,7 +762,7 @@ if __name__ == '__main__':
             self.img = Image.fromstring('RGBA', (3, 3), string)
 
         def renderTile(self, *args, **kwargs):
-            return self.img
+            return SizelessImage(self.img)
 
     def tinybitmap_layer(config, string):
         """ Gin up a fake layer with a TinyBitmap provider.
@@ -877,6 +1018,9 @@ if __name__ == '__main__':
                 # 50% gray all over
                 'gray':       tinybitmap_layer(self.config, _808f * 9),
                 
+                # nothing anywhere
+                'nothing':    tinybitmap_layer(self.config, _0000 * 9),
+                
                 # opaque horizontal gradient, black to white
                 'h gradient': tinybitmap_layer(self.config, (_000f + _808f + _ffff) * 3),
                 
@@ -980,5 +1124,28 @@ if __name__ == '__main__':
             assert img.getpixel((0, 2)) == (0x80, 0x80, 0x80, 0xFF), 'bottom left pixel'
             assert img.getpixel((1, 2)) == (0x40, 0x40, 0x40, 0xFF), 'bottom center pixel'
             assert img.getpixel((2, 2)) == (0x00, 0x00, 0x00, 0xFF), 'bottom right pixel'
+        
+        def test4(self):
+            
+            stack = \
+                [
+                    [
+                        {"src": "nothing"},
+                        {"src": "white wipe"}
+                    ]
+                ]
+            
+            layer = minimal_stack_layer(self.config, stack)
+            img = layer.provider.renderTile(3, 3, None, ModestMaps.Core.Coordinate(0, 0, 0))
+            
+            assert img.getpixel((0, 0)) == (0x00, 0x00, 0x00, 0x00), 'top left pixel'
+            assert img.getpixel((1, 0)) == (0x00, 0x00, 0x00, 0x00), 'top center pixel'
+            assert img.getpixel((2, 0)) == (0x00, 0x00, 0x00, 0x00), 'top right pixel'
+            assert img.getpixel((0, 1)) == (0xFF, 0xFF, 0xFF, 0x80), 'center left pixel'
+            assert img.getpixel((1, 1)) == (0xFF, 0xFF, 0xFF, 0x80), 'middle pixel'
+            assert img.getpixel((2, 1)) == (0xFF, 0xFF, 0xFF, 0x80), 'center right pixel'
+            assert img.getpixel((0, 2)) == (0xFF, 0xFF, 0xFF, 0xFF), 'bottom left pixel'
+            assert img.getpixel((1, 2)) == (0xFF, 0xFF, 0xFF, 0xFF), 'bottom center pixel'
+            assert img.getpixel((2, 2)) == (0xFF, 0xFF, 0xFF, 0xFF), 'bottom right pixel'
     
     unittest.main()
